# Lyapunov_2d_le.py
#
# This code constructs and trains the neural network for Example (7.1) with
# loss function (6.6) in the paper "Computing Lyapunov functions using deep
# neural networks" by Lars Grüne. The code saves the resulting network to
# the file "Lyapunov_2d_le.h5" and plots the resulting Lyapunov function
#
# (C) Lars Grüne 2020


###### import necessary modules ######

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

import tensorflow as tf

from tensorflow import keras

import tensorflow.keras.backend as kb

import matplotlib.pyplot as plt

from mpl_toolkits import mplot3d

import time

###### Auxiliary functions ######

# define the vector field
def vf(x):
    y = [-x[:,0]-10.*x[:,1]**2,-2.*x[:,1]]
    return y

# define the part of the loss function implementing the boundary condition
@tf.function
def bound_loss(y_pred,ubound,lbound,zdata):
    lower_loss = kb.sum(kb.square(kb.min([y_pred-lbound,zdata],0)))
    upper_loss = kb.sum(kb.square(kb.max([y_pred-ubound,zdata],0)))
    custom_loss= lower_loss + upper_loss
    return custom_loss

# define the part of the loss function implementing the PDI
@tf.function
def grad_loss_le(gradx, x_batch_train, vf_batch_train, zdata):
    diff = kb.sum(gradx*vf_batch_train,axis=1) + kb.sum(kb.square(x_batch_train),axis=1)
    g_loss = kb.sum(kb.square(kb.max([diff,zdata],0)))
    return g_loss

# define the upper bound for the boundary condition
def upperbound ( data ):
    return 10.*data[:,0]**2 + 10.*data[:,1]**2

# define the lower bound for the boundary condition
def lowerbound ( data ):
    return 0.1*data[:,0]**2 + 0.1*data[:,1]**2

       
####### Construct the neural network model #######

from tensorflow.keras import layers

time1 = time.perf_counter()

model = tf.keras.Sequential()

# Dimensions and number of subsystems
inputdim = 2
subdim = 1
subnum = 2
sublayersize = 128

# Define inputs:
inputs = keras.Input(shape=(inputdim,), name = 'state')


# define layer for coordinate transformation
xc = []
basename = 'coord_trafo_'

for i in range(subnum):
    thisname = basename + str(i)
    xc.append(layers.Dense(subdim, activation='linear', name = thisname)(inputs))


# define sublayers of dimension sublayersize for approximation of the
# \hat V_i in the compositional form (3.1) of the Lyapunov function
xs = []
basename = 'subsystem_'

for i in range(subnum):
    thisname = basename + str(i)
    xs.append(layers.Dense(sublayersize, activation='softplus', name = thisname)(xc[i]))


# concatenate the sublayers to compute the scalar output W
x = layers.concatenate(xs)

output = layers.Dense(1, activation='linear', name = 'Lyapunov_function')(x)

# compile the model and print summary
model = keras.Model(inputs=inputs, outputs=output)

model.summary()



###### train the model ######


# define the optimizer
optimizer = keras.optimizers.Adam()

# define the training data and parameters for optimization
data = 2.*np.random.random((200000, inputdim)) - 1; 
tdata = tf.convert_to_tensor(data, dtype=tf.float32)

ubound = upperbound(data)
lbound = lowerbound(data)
vfdata = vf(data)

epochs = 20
tol = 1e-5
batch_size = 32
gradweight = 1

tdata   = tf.convert_to_tensor(data,   dtype=tf.float32)
tubound = tf.convert_to_tensor(ubound, dtype=tf.float32)
tlbound = tf.convert_to_tensor(lbound, dtype=tf.float32)
tvf     = kb.transpose(tf.convert_to_tensor(vfdata, dtype=tf.float32))
tzeros  = kb.zeros(tubound.shape)



train_dataset_raw = tf.data.Dataset.from_tensor_slices((tdata, tubound, tlbound, tvf, tzeros))


# this function performs the optimization for one batch of the data
@tf.function
def train_batch(x_batch_train, ub_batch_train, lb_batch_train, vf_batch_train, zeros_batch_train):

# start gradient recording for derivative w.r.t. model parameters
  with tf.GradientTape() as tape:

    # evaluate model
    logits = model(x_batch_train, training=True)

    # evaluate boundary condition part of loss function
    bloss = bound_loss(logits[:,0], ub_batch_train, lb_batch_train, zeros_batch_train)

    # start gradient recording for derivative w.r.t. x
    with tf.GradientTape() as tapex:

      # prepare for evaluation of x-derivative
      tapex.watch(x_batch_train)
      logits2 = model(x_batch_train)

      # evaluate x-derivative
      gradx = tapex.gradient(logits2, x_batch_train)

      # evaluate PDE part of loss function
      gloss = grad_loss_le(gradx, x_batch_train, vf_batch_train, zeros_batch_train )
      loss_value = bloss + gradweight * gloss ### grad_loss_le or grad_loss_eq
      
    # evaluate derivative w.r.t. model parameters
    grads = tape.gradient(loss_value, model.trainable_weights)

    # Run one step of gradient descent optimizer
    optimizer.apply_gradients(zip(grads, model.trainable_weights))

    return(loss_value,bloss,gloss)


# loop through the epochs
for epoch in range(epochs):

  # prepare dataset
  train_dataset = train_dataset_raw.shuffle(buffer_size=1024).batch(batch_size)

  # initialize error variables, mlv=maximum (L_infty), slv = L_1
  mlv = 0.0
  slv = 0.0
  
  # Iterate over the batches of the dataset
  for step, (x_batch_train, ub_batch_train, lb_batch_train, vf_batch_train, zeros_batch_train) in enumerate(train_dataset):

    # call optimization routine
    [loss_value,bloss,gloss] = train_batch(x_batch_train, ub_batch_train, lb_batch_train, vf_batch_train, zeros_batch_train)

    # update errors
    mlv = kb.max([mlv,loss_value],0)
    slv = slv + loss_value
          
    # print log information every 100 steps
    if step % 100 == 0:
      print('epoch %2s, samples %7s, loss %10.6f, aloss %10.6f, mloss %10.6f' % (epoch, ((step + 1) * batch_size), float(loss_value), slv/ (step + 1), mlv), end = '\r')
      
  # print log information at end of epoch    
  print('epoch %2s, samples %7s, loss %10.6f, aloss %10.6f, mloss %10.6f' % (epoch, ((step + 1) * batch_size), float(loss_value), slv/ (step + 1), mlv))
  
  
  # terminate if L_infty error is sufficiently small
  if mlv < tol:
    break



# output computation time
time2 = time.perf_counter()
timediff = time2 - time1
print('time for learning: %fs' % timediff)

# save trained model
model.save('Lyapunov_2d_le.h5')

###### plot result ######

# define resolution
numpoints = 30

# define plotting range and mesh
x = np.linspace(-1, 1, numpoints)
y = np.linspace(-1, 1, numpoints)

X, Y = np.meshgrid(x, y)

s = X.shape

Ze = np.zeros(s)
Zp = np.zeros(s)
DT = np.zeros((numpoints**2,inputdim))

# convert mesh into point vector for which the model can be evaluated
c = 0
for i in range(s[0]):
    for j in range(s[1]):
        DT[c,0] = X[i,j]
        DT[c,1] = Y[i,j]
        c = c+1;

# evaluate model (= Lyapunov function values V)
Ep = model.predict(DT)

# convert point vector to tensor for evaluating x-derivative
tDT = tf.convert_to_tensor(DT, dtype=tf.float32)

# evaluate gradients DV of Lyapunov function
with tf.GradientTape() as tape:
    tape.watch(tDT)
    ypm = model(tDT)
    grads = tape.gradient(ypm, tDT)

# compute orbital derivative DVf
Ee = kb.sum(grads*kb.transpose(tf.convert_to_tensor(vf(DT), dtype=tf.float32)), axis=1)

# copy V and DVf values into plottable format
c = 0
for i in range(s[0]):
    for j in range(s[1]):
        Ze[i,j] = Ee[c]
        Zp[i,j] = Ep[c]
        c = c+1;

# define figure
fig = plt.figure()
ax = plt.axes(projection='3d')

ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z');

# plot values V
ax.plot_surface(X, Y, Zp, rstride=1, cstride=1,
                cmap='viridis', edgecolor='none')

# plot orbital derivative DVf
ax.plot_wireframe(X, Y, Ze, rstride=1, cstride=1)

plt.show()

